//////////////////////////////////////////////////////////////
//
//     IconicFontGlyph
//     Iconic font's character instead of TGlyph
//
//     Author:    Dmitriy Nekrasov
//     E-mail:    dnekrasov@tut.by
//
//     2016-2018
//
//////////////////////////////////////////////////////////////

unit UIExt.IconicFont;

interface

uses
  {$IFDEF MSWINDOWS}
    Winapi.Windows, Winapi.Messages,
  {$ENDIF}

  System.Classes, System.Types, System.UITypes, System.SysUtils, System.IOUtils, System.Rtti, System.UIConsts,

  FMX.Types, FMX.Utils, FMX.Objects, FMX.Controls, FMX.Styles.Objects, FMX.Styles, FMX.Layouts, FMX.Ani, FMX.StdCtrls;

type
  TIconicFont = record
    Short: String;
    Family: String;
    FileName: String;
    Offset: Single;
    CharCount: Integer;
    CharMap: array of TIdentMapEntry;
  public const
      DefFontsFolder = 'Fonts';
      DefFontSize = 24;

      GlyphStyleName = 'IconicFontGlyph';
      NormalColorMask = $CCFFFFFF;
  private
    function CharToIdent(const AChar: Integer; var AIdent: string): Boolean;
    function IdentToChar(const AIdent: String; var AChar: Integer): Boolean;
  public
    procedure ApplyButtonStyle(Sender: TCustomButton; AChar: WideChar; ASize: Single; AColor: TAlphaColor); overload;

    procedure FillCharMap(const ACharMap: array of TIdentMapEntry);

    function CharToString(const AChar: Char): String;
    function StringToChar(AStr: String): Char;
  public
    class procedure ApplyButtonStyle(Sender: TCustomButton); overload; static;

    class function ByFamily(AName: String): TIconicFont; static;
    class function ByFileName(AShort: String): TIconicFont; static;
    class function ByShort(AShort: String): TIconicFont; static;

    class procedure FontsReg(AFontsFolder: String = ''); static;
    class procedure FontsUnreg(AFontsFolder: String = ''); static;
  end;

const
  IF_Empty: TIconicFont = (Short: ''; Family: ''; FileName: ''; Offset: 0);

var
  IconicFonts: array of TIconicFont;

implementation


procedure TIconicFont.ApplyButtonStyle(Sender: TCustomButton; AChar: WideChar; ASize: Single; AColor: TAlphaColor);
var
  txt: TText;
  ifg: TButtonStyleTextObject;

  procedure SetDefColors;
  var
    i: Integer;
  begin
    if txt is TButtonStyleTextObject then
    begin
      ifg.TextSettings.FontColor := txt.TextSettings.FontColor;

      ifg.NormalColor := TButtonStyleTextObject(txt).NormalColor;
      ifg.HotColor := TButtonStyleTextObject(txt).HotColor;
      ifg.FocusedColor := TButtonStyleTextObject(txt).FocusedColor;
      ifg.PressedColor := TButtonStyleTextObject(txt).PressedColor;

      ifg.NormalShadow.Assign(TButtonStyleTextObject(txt).NormalShadow);
      ifg.HotShadow.Assign(TButtonStyleTextObject(txt).HotShadow);
      ifg.FocusedShadow.Assign(TButtonStyleTextObject(txt).FocusedShadow);
      ifg.PressedShadow.Assign(TButtonStyleTextObject(txt).PressedShadow);
    end
    else if (txt is TText) and (txt.ChildrenCount > 0) then
      for i := 0 to txt.ChildrenCount - 1 do
        if (txt.Children[i] is TColorAnimation) then
        begin
          if String(TColorAnimation(txt.Children[i]).Trigger).ToLower.Contains('mouseOwer=true') then
            ifg.HotColor := TColorAnimation(txt.Children[i]).StopValue;
          if String(TColorAnimation(txt.Children[i]).Trigger).ToLower.Contains('ispressed=true') then
            ifg.PressedColor := TColorAnimation(txt.Children[i]).StopValue;
          if String(TColorAnimation(txt.Children[i]).Trigger).ToLower.Contains('isfocused=true') then
            ifg.FocusedColor := TColorAnimation(txt.Children[i]).StopValue;
        end;
  end;

begin
  if Sender.ControlType <> TControlType.Styled then
    Exit;

  Sender.FindStyleResource<TText>('text', txt);
  if not Sender.FindStyleResource<TButtonStyleTextObject>(GlyphStyleName, ifg) then
  begin
    if Assigned(txt) then
    begin
      ifg := TButtonStyleTextObject.Create(txt.Owner);
      ifg.Parent := txt.Parent;
      ifg.StyleName := GlyphStyleName;
      ifg.Align := TAlignLayout.MostLeft;
      ifg.HitTest := False;
      ifg.Width := ifg.Height;
      ifg.ClipParent := True;

      ifg.Margins.Right := 0;
      txt.Margins.Left := 0;

      SetDefColors;

      Sender.StylesData[GlyphStyleName] := TValue.From<TButtonStyleTextObject>(ifg);
    end
    else
      Exit;
  end;

  ifg.Width := ifg.Height;
  ifg.Visible := True;

  if not GlobalUseDirect2D then
    if Assigned(txt) then
      ifg.Margins.Top := txt.Margins.Top + 4 + ASize * Offset
    else
      ifg.Margins.Top := 4 + ASize * Offset
  else if Assigned(txt) then
    ifg.Margins.Top := txt.Margins.Top + ASize * Offset
  else
    ifg.Margins.Top := ASize * Offset;

  if AColor <> TAlphaColorRec.Null then
  begin
    ifg.TextSettings.FontColor := AColor;
    ifg.NormalColor := AColor;
    ifg.HotColor := AColor;
    ifg.FocusedColor := AColor;
    ifg.PressedColor := AColor;
  end
  else if Assigned(txt) then
    SetDefColors;

  ifg.TextSettings.Font.Family := Family;
  ifg.TextSettings.Font.Size := ASize;
  ifg.Text := AChar;
end;

class procedure TIconicFont.ApplyButtonStyle(Sender: TCustomButton);
var
  sStyle: String;
  Font: TIconicFont;
  s: String;
  ch: WideChar;
  cl: TAlphaColor;
  sz: Single;
begin
  try
    sStyle := Sender.StyleName;
    if Sender.StyleName.IsEmpty or (Sender.ControlType <> TControlType.Styled) then
      Exit;

    s := GetToken(sStyle, ';');
    if s.IsEmpty then
      Exit;
    Font := TIconicFont.ByShort(s);
    if Font.Family.IsEmpty then
      Exit;

    ch := Font.StringToChar(GetToken(sStyle, ';'));

    s := GetToken(sStyle, ';');
    if not s.IsEmpty then
      sz := s.ToInteger
    else
      sz := DefFontSize;

    s := GetToken(sStyle, ';');
    if not s.IsEmpty then
      cl := StringToAlphaColor(s)
    else
      cl := TAlphaColorRec.Null;

    Font.ApplyButtonStyle(Sender, ch, sz, cl)
  except
//    on E: Exception do
//      raise Exception.CreateFmt('%s'#$D#$A'%s - %s', [Sender.Name, Sender.StyleName, E.Message]);
  end;
end;

class function TIconicFont.ByFamily(AName: String): TIconicFont;
var
  tif: TIconicFont;
begin
  for tif in IconicFonts do
    if SameText(AName, tif.Family) then
      Exit(tif);

  Result := IF_Empty;
end;

class function TIconicFont.ByFileName(AShort: String): TIconicFont;
var
  tif: TIconicFont;
begin
  for tif in IconicFonts do
    if SameText(AShort, tif.FileName) then
      Exit(tif);

  Result := IF_Empty;
end;

class function TIconicFont.ByShort(AShort: String): TIconicFont;
var
  tif: TIconicFont;
begin
  for tif in IconicFonts do
    if SameText(AShort, tif.Short) then
      Exit(tif);

  Result := IF_Empty;
end;

function TIconicFont.CharToIdent(const AChar: Integer; var AIdent: string): Boolean;
begin
  Result := IntToIdent(AChar, AIdent, CharMap);
  if not Result then
  begin
    AIdent := 'x' + IntToHex(AChar, 4);
    Result := True;
  end;
end;

function TIconicFont.CharToString(const AChar: Char): String;
begin
  CharToIdent(Ord(AChar), Result);
  if Result.Chars[0] = 'x' then
    Result := '#' + Result.SubString(1);
end;

function TIconicFont.IdentToChar(const AIdent: String; var AChar: Integer): Boolean;
begin
  if (AIdent.Length > 0) and (AIdent.Chars[0] = 'x') then
  begin
    AChar := Ord(StringToChar(AIdent));
    Result := True;
  end else
    Result := IdentToInt(AIdent, AChar, CharMap);
  if not Result and (AIdent.Length > 0) then
  try
    AChar := AIdent.ToInteger;
    Result := True;
  except
    Result := False;
  end;
end;

function TIconicFont.StringToChar(AStr: String): Char;
var
  iChar: Integer;
begin
  if AStr.IsEmpty then
    Exit(#0)
  else if (AStr.Chars[0] = '#') or (AStr.Chars[0] = 'x') then
    AStr := '$' + AStr.SubString(1);

  if not IdentToChar(AStr, iChar) then
    Result := Char(StrToInt(AStr))
  else
    Result := Char(iChar);
end;

procedure TIconicFont.FillCharMap(const ACharMap: array of TIdentMapEntry);
var
  i: Integer;
begin
  CharCount := Length(ACharMap);
  SetLength(CharMap, CharCount);
  for i := Low(ACharMap) to High(ACharMap) do
    CharMap[i] := ACharMap[i];
end;

class procedure TIconicFont.FontsReg(AFontsFolder: String);
{$IFDEF MSWINDOWS}
var
  tif: TIconicFont;
  sPath, sFile: String;
  rs: TResourceStream;
{$ENDIF}
begin
{$IFDEF MSWINDOWS}
  if AFontsFolder.IsEmpty then
    sPath := System.IOUtils.TPath.Combine(System.IOUtils.TPath.GetLibraryPath, DefFontsFolder)
  else
    sPath := IncludeTrailingPathDelimiter(AFontsFolder);

  for tif in IconicFonts do
  begin
    sFile := System.IOUtils.TPath.Combine(sPath, tif.FileName);
    if not System.IOUtils.TFile.Exists(sFile) then
    try
      ForceDirectories(sPath);
      rs := TResourceStream.Create(HInstance, Format('IF_%s', [tif.Short]), RT_RCData);
      try
        rs.SaveToFile(sFile);
      finally
        rs.Free;
      end;
    except
    end;
    if System.IOUtils.TFile.Exists(sFile) then
      AddFontResourceEx(PWideChar(sFile), FR_NOT_ENUM, nil);
  end;
{$ENDIF}
end;

class procedure TIconicFont.FontsUnreg(AFontsFolder: String);
{$IFDEF MSWINDOWS}
var
  tif: TIconicFont;
  sPath, sFile: String;
{$ENDIF}
begin
{$IFDEF MSWINDOWS}
  if AFontsFolder.IsEmpty then
    sPath := System.IOUtils.TPath.Combine(System.IOUtils.TPath.GetLibraryPath, DefFontsFolder)
  else
    sPath := IncludeTrailingPathDelimiter(AFontsFolder);

  for tif in IconicFonts do
  begin
    sFile := System.IOUtils.TPath.Combine(sPath, tif.FileName);
    if System.IOUtils.TFile.Exists(sFile) then
      RemoveFontResourceEx(PWideChar(sFile), FR_NOT_ENUM, nil);
  end;
{$ENDIF}
end;

end.
